import numpy as np
import mpmath as mp
import birdepy.utility as ut


def continued_fraction_terms_00_mp(idx, s, param, b_rate, d_rate):
    """
    Returns a partial numerator and denominator for the Laplace transform of the
    transition probability from 0 to 0 of a population-size-dependent
    birth-and-death process (as defined as $a_n$ and $b_n$ below Equation (9) in
    reference [1]).

    [1] Crawford, F. W., & Suchard, M. A. (2012). Transition probabilities for
    general birth–death processes with applications in ecology, genetics, and
    evolution. Journal of mathematical biology, 65(3), 553-580.
    """
    if idx == 0:
        return mp.mpf('0.0'), mp.mpf('0.0')
    elif idx == 1:
        return mp.mpf('1.0'), mp.mpmathify(s + b_rate(0, param))
    else:
        return mp.fmul(-b_rate(idx - 2, param), d_rate(idx - 1, param)), \
               mp.fsum((s, b_rate(idx - 1, param), d_rate(idx - 1, param)))


def big_b_mp(i, j, s, param, b_rate, d_rate):
    """
    Returns the denominator of the 'j', 'i', and 'i+1' convergent (approximant)
    of the Laplace transform of the transition probability from 0 to 0 of a
    population-size-dependent birth-and-death process (as defined as $B_k$ in
    Lemma 1 of reference [1]).

    [1] Crawford, F. W., & Suchard, M. A. (2012). Transition probabilities for
    general birth–death processes with applications in ecology, genetics, and
    evolution. Journal of mathematical biology, 65(3), 553-580.
    """
    # This function assumes i >= j and returns [B_j, B_i, B_{i+1}]
    if i == 1:  # return B0, B1, B2
        B_0 = mp.mpf('1.0')
        B_1 = mp.fadd(mp.mpmathify(s), mp.mpmathify(b_rate(0, param)))
        B_2 = mp.fsub(
            mp.fmul(mp.fadd(s + mp.mpmathify(b_rate(1, param)),
                            mp.mpmathify(d_rate(1, param))),
                    B_1),
            mp.fprod((mp.mpmathify(b_rate(0, param)),
                      mp.mpmathify(d_rate(1, param)),
                      B_0))
        )
        if j == 0:
            return B_0, B_1, B_2
        elif j == 1:
            return B_1, B_1, B_2
    else:
        cv = [1,
              mp.fadd(s, mp.mpmathify(b_rate(0, param)))]  # [B0, B1]
        if j == 0:
            B_j = cv[0]
        elif j == 1:
            B_j = cv[1]
        for terms in np.arange(2, i + 2, 1):
            cv = [cv[1],
                  mp.fsub(
                      mp.fmul(
                          mp.fsum((s,
                                   mp.mpmathify(
                                       b_rate(terms - 1, param)),
                                   mp.mpmathify(
                                       d_rate(terms - 1, param)))
                                  ),
                          cv[1]),
                      mp.fprod((mp.mpmathify(b_rate(terms - 2, param)),
                                mp.mpmathify(d_rate(terms - 1, param)),
                                cv[0])))
                  ]
            if terms == j:
                B_j = cv[1]
        return B_j, cv[0], cv[1]


def mu_product(i, j, param, d_rate):
    """
    Returns a product of the death rates from 'j+1' to 'i' as used in
    Theorem 1 of reference [1].

    [1] Crawford, F. W., & Suchard, M. A. (2012). Transition probabilities for
    general birth–death processes with applications in ecology, genetics, and
    evolution. Journal of mathematical biology, 65(3), 553-580.
    """
    return np.prod(d_rate(np.arange(j + 1, i + 1, 1), param))


def lam_product(i, j, param, b_rate):
    """
    Returns a product of the birth rates from 'i' to 'j-1' as used in
    Theorem 1 of reference [1].

    [1] Crawford, F. W., & Suchard, M. A. (2012). Transition probabilities for
    general birth–death processes with applications in ecology, genetics, and
    evolution. Journal of mathematical biology, 65(3), 553-580.
    """
    return np.prod(b_rate(np.arange(i, j, 1), param))


def mod_lentz_00_mp(fun, eps):
    """
    Computes part of the Laplace transform of the transition probability from
    0 to 0 of a population-size-dependent birth-and-death process (as defined
    by Equation (15) in reference [1]) in high precision using the modified
    Lentz algorithm.

    [1] Crawford, F. W., & Suchard, M. A. (2012). Transition probabilities for
    general birth–death processes with applications in ecology, genetics, and
    evolution. Journal of mathematical biology, 65(3), 553-580.
    """
    (a0, b0) = fun(0)
    f = mp.mpmathify(b0)
    if f == 0:
        f = mp.mpf('1e-30')
    C = f
    D = 0
    error = 1 + eps
    idx = 0
    while error > eps:
        idx += 1
        (a, b) = fun(idx)
        D = mp.fadd(b, mp.fmul(a, D))
        if D == 0:
            D = mp.mpf('1e-30')
        C = mp.fadd(b, mp.fdiv(a, C))
        if C == 0:
            C = mp.mpf('1e-30')
        D = mp.fdiv(1.0, D)
        delta = mp.fmul(C, D)
        f = mp.fmul(f, delta)
        error = mp.fabs(mp.fsub(delta, 1.0))
    return f


def mod_lentz_ij_mp(s, i, j, param, b_rate, d_rate, eps):
    """
    Computes part of the Laplace transform of the transition probability from
    i to j of a population-size-dependent birth-and-death process (as defined
    by Equation (15) in reference [1]) in high precision using the modified
    Lentz algorithm.

    [1] Crawford, F. W., & Suchard, M. A. (2012). Transition probabilities for
    general birth–death processes with applications in ecology, genetics, and
    evolution. Journal of mathematical biology, 65(3), 553-580.
    """
    B_j, B_i, B_ip1 = big_b_mp(i, j, s, param, b_rate, d_rate)
    f = mp.mpf('1e-30')
    C = f
    a1, b1 = mp.mpmathify(B_j), mp.mpmathify(B_ip1)
    D = b1
    if D == 0:
        D = mp.mpf('1e-30')
    C = mp.fadd(b1, mp.fdiv(a1, C))
    if C == 0:
        C = mp.mpf('1e-30')
    D = mp.fdiv(mp.mpf('1.0'), D)
    delta = mp.fmul(C, D)
    f = mp.fmul(f, delta)
    a2, b2 = mp.fprod((B_i,
                       -mp.mpmathify(b_rate(i, param)),
                       mp.mpmathify(d_rate(i + 1, param)))), \
             mp.fsum((s,
                      mp.mpmathify(b_rate(i + 1, param)),
                      mp.mpmathify(d_rate(i + 1, param))))
    D = mp.fadd(b2, mp.fmul(a2, D))
    if D == 0:
        D = mp.mpf('1e-30')
    C = mp.fadd(b2, mp.fdiv(a2, C))
    if C == 0:
        C = mp.mpf('1e-30')
    D = mp.fdiv(mp.mpf('1.0'), D)
    delta = mp.fmul(C, D)
    f = mp.fmul(f, delta)
    error = mp.fabs(mp.fsub(delta, mp.mpf('1.0')))
    idx = 3
    while error > eps:  # & (idx < 1000):
        (a, b) = mp.fmul(-mp.mpmathify(b_rate(i + idx - 2, param)),
                         mp.mpmathify(d_rate(i + idx - 1, param))), \
                 mp.fsum((s,
                          mp.mpmathify(b_rate(i + idx - 1, param)),
                          mp.mpmathify(d_rate(i + idx - 1, param))))
        D = mp.fadd(b, mp.fmul(a, D))
        if D == 0:
            D = mp.mpf('1e-30')
        C = mp.fadd(b, mp.fdiv(a, C))
        if C == 0:
            C = mp.mpf('1e-30')
        D = mp.fdiv(mp.mpf('1.0'), D)
        delta = mp.fmul(C, D)
        f = mp.fmul(f, delta)
        error = mp.fabs(mp.fsub(delta, mp.mpf('1.0')))
        idx += 1
    return f


def laplace_p(s, i, j, param, b_rate, d_rate, eps):
    """
    Computes the Laplace transform of the transition probability from
    'i' to 'j' of a population-size-dependent birth-and-death process (as defined
    by Equation (15) in reference [1]) in high precision using the modified
    Lentz algorithm.

    [1] Crawford, F. W., & Suchard, M. A. (2012). Transition probabilities for
    general birth–death processes with applications in ecology, genetics, and
    evolution. Journal of mathematical biology, 65(3), 553-580.
    """
    if i == 0 and j == 0:
        return mod_lentz_00_mp(
            lambda idx:
            continued_fraction_terms_00_mp(idx, s, param, b_rate, d_rate),
            eps)
    elif i >= j:
        return mp.fmul(mu_product(i, j, param, d_rate),
                       mod_lentz_ij_mp(s, i, j, param, b_rate, d_rate, eps))
    else:
        return mp.fmul(lam_product(i, j, param, b_rate),
                       mod_lentz_ij_mp(s, j, i, param, b_rate, d_rate, eps))


def probability_ilt(z0, zt, t, param, b_rate, d_rate, lentz_eps, laplace_method, k):
    """Transition probabilities for continuous-time birth-and-death processes
    using the *inverse Laplace transform* method.

    To use this function call :func:`birdepy.probability` with `method` set to
    'ilt'::

        birdepy.probability(z0, zt, t, param, method='ilt', eps=1e-6,
                            laplace_method='cme-talbot', precision=100, k=25)

    The parameters associated with this method (listed below) can be
    accessed using kwargs in :func:`birdepy.probability()`. See documentation
    of :func:`birdepy.probability` for the main arguments.

    Parameters
    ----------
    lentz_eps : scalar, optional
        Termination threshold for Lentz algorithm computation of Laplace
        domain functions.

    laplace_method : string, optional
        Numerical inverse Laplace transform algorithm to use. Should be one of:
        'cme-talbot' (default), 'cme', 'euler', 'gaver', 'talbot', 'stehfest',
        'dehoog', 'cme-mp' or 'gwr'.

    precision : int, optional
        Numerical precision (only used for methods that invoke mpmath).

    k: int, optional
        Maximum number of terms used for Laplace transform numerical inversion.
        Only applicable if argument 'laplace_method' is set to 'cme-talbot',
        'cme', 'euler', 'gaver' or 'cme-mp'.
        See https://www.inverselaplace.org/ for more information.

    Examples
    --------
    Approximate transition probability for a Verhulst model using numerical inverse Laplace transform: ::

        import birdepy as bd
        bd.probability(19, 27, 1.0, [0.5, 0.3, 0.02, 0.01], model='Verhulst', method='ilt')[0][0]

    Outputs: ::

        0.0027403264310572615

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    For more details on this method see [3].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    .. [3] Crawford, F.W. and Suchard, M.A., 2012. Transition probabilities
     for general birth-death processes with applications in ecology, genetics,
     and evolution. Journal of Mathematical Biology, 65(3), pp.553-580.

    """
    # If more than one time is requested it is easiest to divert into a different code block
    if t.size == 1:
        # Initialize an array to store output
        output = np.zeros((z0.size, zt.size))
        for idx1, _z0 in enumerate(z0):
            for idx2, _zt in enumerate(zt):
                # Compute and invert Laplace transform of probability, and ensure a float is
                # returned
                pr = float(ut.laplace_invert(
                    lambda s: laplace_p(s, _z0, _zt, param, b_rate, d_rate,
                                        lentz_eps),
                    t[0], laplace_method=laplace_method, k=k,
                    f_bounds=[0.0, 1.0]))
                output[idx1, idx2] = pr
    else:
        # Initialize an array to store output
        output = np.zeros((t.size, z0.size, zt.size))
        for idx1, _z0 in enumerate(z0):
            for idx2, _zt in enumerate(zt):
                for idx3, _t in enumerate(t):
                    # Compute and invert Laplace transform of probability, and ensure a float is
                    # returned
                    pr = float(ut.laplace_invert(
                        lambda s: laplace_p(s, _z0, _zt, param, b_rate, d_rate,
                                            lentz_eps),
                        _t, laplace_method=laplace_method, k=k,
                        f_bounds=[0.0, 1.0]))
                    output[idx3, idx1, idx2] = pr
    return output



import numpy as np
import birdepy.utility as ut
from birdepy.interface_aug import f_fun_bld as f_fun_bld
from birdepy.interface_dnm import bld_ll_fun
import birdepy.utility_em as utem


def discrete_est_em_cg(data, sorted_data, p0, likelihood, technique, known_p,
                       idx_known_p, model, b_rate, d_rate, z_trunc, p_bounds,
                       con, max_it, i_tol, j_tol, h_tol, display, opt_method,
                       options):
    """
    Executes an expectation-maximization algorithm, which is accelerated using
    conjugate gradient ideas, to estimate parameters for a
    population-size-dependent birth-and-death process.

    This function is called by :func:`bd.interface_aug.f_fun_bld()`.

    [1] Jamshidian, M., & Jennrich, R. I. (1997). Acceleration of the EM algorithm
     by using quasi‐Newton methods. Journal of the Royal Statistical Society:
     Series B (Statistical Methodology), 59(3), 569-587.

    [2] Jamshidian, M., & Jennrich, R. I. (1993). Conjugate gradient acceleration of
     the EM algorithm. Journal of the American Statistical Association, 88(421), 221-228.
    """
    #Initialize a list to return parameters in each iteration and store the initial parameters in it
    iterations = [p0]

    # Build the log-likelihood function assuming all parameters are known
    pre_ll = bld_ll_fun(data, likelihood, model, z_trunc, options)

    # Adjust the log-likelihood function to account for unknown parameters and
    # negate it so that the minimize function performs maximisation
    def ll(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return pre_ll(param)

    it = 0
    p_est_it_m1 = p0
    difference = i_tol + 1
    # Iterate until the difference between parameter estimates is smaller
    # than i_tol or max_it iterations have been performed
    while difference > i_tol and it < max_it:

        # Build surrogate log-likelihood (defined in Section 2.1 of ref [1])
        f_fun = f_fun_bld(sorted_data, p_est_it_m1, b_rate, d_rate, likelihood,
                          technique, idx_known_p, known_p, model, z_trunc,
                          j_tol, h_tol, options)[0]

        # Negate the surrogate log-likelihood so that scipy.optimize.minimize()
        # Can be used to find a maximum
        def error_fun(p_prop):
            return -f_fun(p_prop)

        # Find a zero of an EM step (defined as $\tilde g(\theta) at start of Section 2.2
        # of reference [1])
        g_tilde = ut.minimize_(error_fun, p_est_it_m1, p_bounds, con,
                               opt_method, options).x - p_est_it_m1

        # Reset the algorithm whenever the iteration index is an integer multiple of p.
        # This is discussed at the end of Section 2 in reference [2].
        if (it % p0.size) == 0:
            direction = g_tilde
        else:
            # Jacobian of log-likelihood at previous parameter estimate
            J_ell_it_m1 = ut.Jacobian(ll, p_est_it_m1, p_bounds)
            # Jacobian of log-likelihood at parameter estimate before previous parameter
            # estimate
            J_ell_it_m2 = ut.Jacobian(ll, p_est_it_m2, p_bounds)

            # Store current direction since new direction is in terms of current direction
            previous_direction = direction

            # Helps us to choose direction to change parameters (Equation (6) in reference [2])
            beta = np.divide(np.inner(g_tilde,
                                      np.subtract(J_ell_it_m1, J_ell_it_m2)),
                             np.inner(previous_direction,
                                      np.subtract(J_ell_it_m1, J_ell_it_m2)))
            # Tells us direction to change parameters in (Equation (7) in reference [2])
            direction = g_tilde - np.multiply(beta, previous_direction)

        # Need to store this since each iteration depends on previous 2 iterations (except
        # when a reset happens)
        p_est_it_m2 = p_est_it_m1

        # Tells us how much to change parameters by (Equation (4) in reference [2])
        a_opt = utem.line_search_ll(direction, data, p_est_it_m1, likelihood,
                                    known_p, idx_known_p, model, z_trunc,
                                    p_bounds, con, opt_method, options)[0]

        # Update parameter estimate
        p_est_it_m1 = p_est_it_m1 + np.multiply(a_opt, direction)

        # Compute difference to determine if should terminate
        difference = np.sum(np.abs(p_est_it_m1 - p_est_it_m2))
        # Store estimate
        iterations.append(p_est_it_m1)

        # Update iteration number
        it += 1
        # Print a progress update if requested
        if display:
            print('Iteration ', it, ' estimate is: ', p_est_it_m1)

    return list(p_est_it_m1), np.array(iterations)



import numpy as np
from birdepy import simulate
from collections import Counter


def probability_sim(z0, zt, t, param, b_rate, d_rate, k, sim_method, tau, seed):
    """Transition probabilities for continuous-time birth-and-death processes
    using crude Monte Carlo simulation.

    To use this function call :func:`birdepy.probability` with `method` set to
    'sim'::

        birdepy.probability(z0, zt, t, param, method='sim', k=10**5, seed=None)

    The parameters associated with this method (listed below) can be
    accessed using kwargs in :func:`birdepy.probability()`. See documentation
    of :func:`birdepy.probability` for the main arguments.

    Parameters
    ----------

    k: int, optional
        Number of samples used to generate each probability estimate.
        The total number of samples used will be z0.size * k.

    seed : int, Generator, optional
        If *seed* is not specified the random numbers are generated according
        to *np.random.default_rng()*. If *seed* is an *int*, random numbers are
        generated according to *np.random.default_rng(seed)*. If seed is a
        *Generator*, then that object is used. See
        :ref:`here <Reproducibility>` for more information.

    sim_method : string, optional
        Simulation algorithm used to generate samples (see
        :ref:`here<Simulation Algorithms>`). Should be one of:

            - 'exact' (default)
            - 'ea'
            - 'ma'
            - 'gwa'

    Examples
    --------
    >>> import birdepy as bd
    >>> bd.probability(19, 27, 1.0, [0.5, 0.3, 0.02, 0.01], model='Verhulst', method='sim', k=10**5,
    ...                seed=2021)[0][0]
    0.00294

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    For more details on this method see [3].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S. and Patch, B. BirDePy: Parameter estimation for
     population-size-dependent birth-and-death processes in Python. ArXiV, 2021.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    .. [3] Crawford, F.W. and Suchard, M.A., 2012. Transition probabilities
     for general birth-death processes with applications in ecology, genetics,
     and evolution. Journal of Mathematical Biology, 65(3), pp.553-580.

    """

    if t.size == 1:
        output = np.zeros((z0.size, zt.size))
        for idx1, _z0 in enumerate(z0):
            counts = Counter(simulate.discrete(param, 'custom', _z0, [0, t[0]],
                                               b_rate=b_rate, d_rate=d_rate,
                                               k=k, seed=seed,
                                               method=sim_method,
                                               tau=tau)[:, 1])
            for idx2, _zt in enumerate(zt):
                output[idx1, idx2] = counts[_zt]/k
    else:
        output = np.zeros((t.size, z0.size, zt.size))
        for idx3, _t in enumerate(t):
            for idx1, _z0 in enumerate(z0):
                counts = Counter(simulate.discrete(param, 'custom', _z0,
                                                   [0, _t], b_rate=b_rate,
                                                   d_rate=d_rate, k=k,
                                                   seed=seed,
                                                   method=sim_method,
                                                   tau=tau)[:, 1])
                for idx2, _zt in enumerate(zt):
                    output[idx3, idx1, idx2] = counts[_zt] / k
    return output
